import json
import logging
import logging.config
import os
import platform

from colorama import Fore, Style
from colorama import init as colorama_init

from .utils import _resolve_environment, _logging_extras, _logging_prefix


"""
The configuration module hols generic logging configuration utilities.
"""


colorama_init()

if platform.system() == "Windows":
    os.system("color")  # required for colors to work in Windows
logger = logging.getLogger(__name__)


_default_colors = {
    "app": Fore.BLUE,
    "server": Fore.YELLOW,
    "api": Fore.YELLOW,
    "worker": Fore.GREEN,
    "tasks": Fore.GREEN,
    "celery": Fore.GREEN,
    "db": Fore.MAGENTA,
    "psql": Fore.MAGENTA,
    "postgres": Fore.MAGENTA,
}


def is_gunicorn() -> bool:
    return "gunicorn" in os.environ.get("SERVER_SOFTWARE", "")


def configure_prefix(name: str | None = None, color: str | None = None):
    """
    Configure a logging prefix with a name and (optionally) a color and environment. For example, if the name was
    "api" and the environment is "azure", the log would be something like,

    api|azure|INFO: This is a log message.

    where "api|azure" is yellow, unless another color is specified.
    """
    if name is None:
        return
    if color is None:
        color = _default_colors.get(name, "")  # type: ignore
    else:
        color = getattr(Fore, color.upper())
    prefix = f"{color}{name}|{_resolve_environment()}{Style.RESET_ALL}|"
    _logging_extras[_logging_prefix] = prefix


def configure_logging(config_path: str = "logging_config.json"):
    """
    Running in Gunicorn, the logging configuration is already set up. In other cases, the configuration is loaded from the config_path.
    """
    if is_gunicorn():
        return
    with open(config_path, "r") as f:
        config = json.load(f)
    logging.config.dictConfig(config)


def configure_logging_default(
    config_path: str = "logging_config.json",
    name: str | None = None,
    color: str | None = None,
):
    """
    A single entry point for logging configuration that invokes lower level functions with sane defaults.
    """
    configure_logging(config_path)
    configure_prefix(name, color)



"""Core functionality shared by all JSON loggers"""

### IMPORTS
### ============================================================================
## Future
from __future__ import annotations

## Standard Library
from datetime import datetime, timezone
import importlib
import logging
import re
import sys
from typing import Optional, Union, Callable, List, Dict, Container, Any, Sequence

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

## Installed

## Application


### CONSTANTS
### ============================================================================
RESERVED_ATTRS: List[str] = [
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
]
"""Default reserved attributes.

These come from the [default attributes of `LogRecord` objects](http://docs.python.org/library/logging.html#logrecord-attributes).

Note:
    Although considered a constant, this list is dependent on the Python version due to
    different `LogRecord` objects having different attributes in different Python versions.

*Changed in 3.0*: `RESERVED_ATTRS` is now `list[str]` instead of `tuple[str, ...]`.
"""

if sys.version_info >= (3, 12):
    # taskName added in python 3.12
    RESERVED_ATTRS.append("taskName")
    RESERVED_ATTRS.sort()


STYLE_STRING_TEMPLATE_REGEX = re.compile(r"\$\{(.+?)\}", re.IGNORECASE)  # $ style
STYLE_STRING_FORMAT_REGEX = re.compile(r"\{(.+?)\}", re.IGNORECASE)  # { style
STYLE_PERCENT_REGEX = re.compile(r"%\((.+?)\)", re.IGNORECASE)  # % style

## Type Aliases
## -----------------------------------------------------------------------------
OptionalCallableOrStr: TypeAlias = Optional[Union[Callable, str]]
"""Type alias"""

LogRecord: TypeAlias = Dict[str, Any]
"""Type alias"""


### FUNCTIONS
### ============================================================================
def str_to_object(obj: Any) -> Any:
    """Import strings to an object, leaving non-strings as-is.

    Args:
        obj: the object or string to process

    *New in 3.1*
    """

    if not isinstance(obj, str):
        return obj

    module_name, attribute_name = obj.rsplit(".", 1)
    return getattr(importlib.import_module(module_name), attribute_name)


def merge_record_extra(
    record: logging.LogRecord,
    target: Dict,
    reserved: Container[str],
    rename_fields: Optional[Dict[str, str]] = None,
) -> Dict:
    """
    Merges extra attributes from LogRecord object into target dictionary

    Args:
        record: logging.LogRecord
        target: dict to update
        reserved: dict or list with reserved keys to skip
        rename_fields: an optional dict, used to rename field names in the output.
            e.g. Rename `levelname` to `log.level`: `{'levelname': 'log.level'}`

    *Changed in 3.1*: `reserved` is now `Container[str]`.
    """
    if rename_fields is None:
        rename_fields = {}
    for key, value in record.__dict__.items():
        # this allows to have numeric keys
        if key not in reserved and not (hasattr(key, "startswith") and key.startswith("_")):
            target[rename_fields.get(key, key)] = value
    return target


### CLASSES
### ============================================================================
class BaseJsonFormatter(logging.Formatter):
    """Base class for all formatters

    Must not be used directly.

    *New in 3.1*

    *Changed in 3.2*: `defaults` argument is no longer ignored.
    """

    _style: Union[logging.PercentStyle, str]  # type: ignore[assignment]

    ## Parent Methods
    ## -------------------------------------------------------------------------
    # pylint: disable=too-many-arguments,super-init-not-called
    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: str = "%",
        validate: bool = True,
        *,
        prefix: str = "",
        rename_fields: Optional[Dict[str, str]] = None,
        rename_fields_keep_missing: bool = False,
        static_fields: Optional[Dict[str, Any]] = None,
        reserved_attrs: Optional[Sequence[str]] = None,
        timestamp: Union[bool, str] = False,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Args:
            fmt: string representing fields to log
            datefmt: format to use when formatting `asctime` field
            style: how to extract log fields from `fmt`
            validate: validate `fmt` against style, if implementing a custom `style` you
                must set this to `False`.
            defaults: a dictionary containing default fields that are added before all other fields and
                may be overridden. The supplied fields are still subject to `rename_fields`.
            prefix: an optional string prefix added at the beginning of
                the formatted string
            rename_fields: an optional dict, used to rename field names in the output.
                Rename `message` to `@message`: `{'message': '@message'}`
            rename_fields_keep_missing: When renaming fields, include missing fields in the output.
            static_fields: an optional dict, used to add fields with static values to all logs
            reserved_attrs: an optional list of fields that will be skipped when
                outputting json log record. Defaults to [all log record attributes][pythonjsonlogger.core.RESERVED_ATTRS].
            timestamp: an optional string/boolean field to add a timestamp when
                outputting the json log record. If string is passed, timestamp will be added
                to log record using string as key. If True boolean is passed, timestamp key
                will be "timestamp". Defaults to False/off.

        *Changed in 3.1*:

        - you can now use custom values for style by setting validate to `False`.
          The value is stored in `self._style` as a string. The `parse` method will need to be
          overridden in order to support the new style.
        - Renaming fields now preserves the order that fields were added in and avoids adding
          missing fields. The original behaviour, missing fields have a value of `None`, is still
          available by setting `rename_fields_keep_missing` to `True`.
        """
        ## logging.Formatter compatibility
        ## ---------------------------------------------------------------------
        # Note: validate added in 3.8, defaults added in 3.10
        if style in logging._STYLES:
            _style = logging._STYLES[style][0](fmt)  # type: ignore[operator]
            if validate:
                _style.validate()
            self._style = _style
            self._fmt = _style._fmt

        elif not validate:
            self._style = style
            self._fmt = fmt

        else:
            raise ValueError(f"Style must be one of: {','.join(logging._STYLES.keys())}")

        self.datefmt = datefmt

        ## JSON Logging specific
        ## ---------------------------------------------------------------------
        self.prefix = prefix
        self.rename_fields = rename_fields if rename_fields is not None else {}
        self.rename_fields_keep_missing = rename_fields_keep_missing
        self.static_fields = static_fields if static_fields is not None else {}
        self.reserved_attrs = set(reserved_attrs if reserved_attrs is not None else RESERVED_ATTRS)
        self.timestamp = timestamp

        self._required_fields = self.parse()
        self._skip_fields = set(self._required_fields)
        self._skip_fields.update(self.reserved_attrs)
        self.defaults = defaults if defaults is not None else {}
        return

    def format(self, record: logging.LogRecord) -> str:
        """Formats a log record and serializes to json

        Args:
            record: the record to format
        """
        message_dict: Dict[str, Any] = {}
        # TODO: logging.LogRecord.msg and logging.LogRecord.message in typeshed
        #        are always type of str. We shouldn't need to override that.
        if isinstance(record.msg, dict):
            message_dict = record.msg
            record.message = ""
        else:
            record.message = record.getMessage()

        # only format time if needed
        if "asctime" in self._required_fields:
            record.asctime = self.formatTime(record, self.datefmt)

        # Display formatted exception, but allow overriding it in the
        # user-supplied dict.
        if record.exc_info and not message_dict.get("exc_info"):
            message_dict["exc_info"] = self.formatException(record.exc_info)
        if not message_dict.get("exc_info") and record.exc_text:
            message_dict["exc_info"] = record.exc_text

        # Display formatted record of stack frames
        # default format is a string returned from :func:`traceback.print_stack`
        if record.stack_info and not message_dict.get("stack_info"):
            message_dict["stack_info"] = self.formatStack(record.stack_info)

        log_record: LogRecord = {}
        self.add_fields(log_record, record, message_dict)
        log_record = self.process_log_record(log_record)

        return self.serialize_log_record(log_record)

    ## JSON Formatter Specific Methods
    ## -------------------------------------------------------------------------
    def parse(self) -> List[str]:
        """Parses format string looking for substitutions

        This method is responsible for returning a list of fields (as strings)
        to include in all log messages.

        You can support custom styles by overriding this method.

        Returns:
            list of fields to be extracted and serialized
        """
        if isinstance(self._style, logging.StringTemplateStyle):
            formatter_style_pattern = STYLE_STRING_TEMPLATE_REGEX

        elif isinstance(self._style, logging.StrFormatStyle):
            formatter_style_pattern = STYLE_STRING_FORMAT_REGEX

        elif isinstance(self._style, logging.PercentStyle):
            # PercentStyle is parent class of StringTemplateStyle and StrFormatStyle
            # so it must be checked last.
            formatter_style_pattern = STYLE_PERCENT_REGEX

        else:
            raise ValueError(f"Style {self._style!r} is not supported")

        if self._fmt:
            return formatter_style_pattern.findall(self._fmt)

        return []

    def serialize_log_record(self, log_record: LogRecord) -> str:
        """Returns the final representation of the log record.

        Args:
            log_record: the log record
        """
        return self.prefix + self.jsonify_log_record(log_record)

    def add_fields(
        self,
        log_record: Dict[str, Any],
        record: logging.LogRecord,
        message_dict: Dict[str, Any],
    ) -> None:
        """Extract fields from a LogRecord for logging

        This method can be overridden to implement custom logic for adding fields.

        Args:
            log_record: data that will be logged
            record: the record to extract data from
            message_dict: dictionary that was logged instead of a message. e.g
                `logger.info({"is_this_message_dict": True})`
        """
        for field in self.defaults:
            log_record[self._get_rename(field)] = self.defaults[field]

        for field in self._required_fields:
            log_record[self._get_rename(field)] = record.__dict__.get(field)

        for data_dict in [self.static_fields, message_dict]:
            for key, value in data_dict.items():
                log_record[self._get_rename(key)] = value

        merge_record_extra(
            record,
            log_record,
            reserved=self._skip_fields,
            rename_fields=self.rename_fields,
        )

        if self.timestamp:
            key = self.timestamp if isinstance(self.timestamp, str) else "timestamp"
            log_record[self._get_rename(key)] = datetime.fromtimestamp(
                record.created, tz=timezone.utc
            )

        if self.rename_fields_keep_missing:
            for field in self.rename_fields.values():
                if field not in log_record:
                    log_record[field] = None
        return

    def _get_rename(self, key: str) -> str:
        return self.rename_fields.get(key, key)

    # Child Methods
    # ..........................................................................
    def jsonify_log_record(self, log_record: LogRecord) -> str:
        """Convert this log record into a JSON string.

        Child classes MUST override this method.

        Args:
            log_record: the data to serialize
        """
        raise NotImplementedError()

    def process_log_record(self, log_record: LogRecord) -> LogRecord:
        """Custom processing of the log record.

        Child classes can override this method to alter the log record before it
        is serialized.

        Args:
            log_record: incoming data
        """
        return log_record



from inspari.logging.configuration import (
    is_gunicorn,
    configure_logging,
    configure_logging_default,
)
from inspari.logging.streamlogs import AzureBlobStorageHandler
from inspari.logging.pythonjsonlogger.json import JsonFormatter



import os

"""
Utility functions for shared across logging modules.
"""

_logging_prefix = "prefix"
_logging_suffix = "suffix"
_logging_extras: dict[str, str] = {
    _logging_prefix: "",
    _logging_suffix: "",
}


def _resolve_environment() -> str:
    if "WEBSITE_SKU" in os.environ:
        return "azure"
    return "local"



### IMPORTS
### ============================================================================
## Future

## Standard Library
import warnings

## Installed

## Application
from . import json
from . import utils

### CONSTANTS
### ============================================================================
ORJSON_AVAILABLE = utils.package_is_available("orjson")
MSGSPEC_AVAILABLE = utils.package_is_available("msgspec")


### DEPRECATED COMPATIBILITY
### ============================================================================
def __getattr__(name: str):
    if name == "jsonlogger":
        warnings.warn(
            "pythonjsonlogger.jsonlogger has been moved to pythonjsonlogger.json",
            DeprecationWarning,
        )
        return json
    raise AttributeError(f"module {__name__} has no attribute {name}")



import logging
import logging.config
import os
import time
import traceback
from datetime import UTC, datetime
from typing import Optional

import typer
from azure.core.exceptions import HttpResponseError, ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobClient, BlobServiceClient, ContentSettings
from pydantic import Field

from inspari.config import load_dotenv
from inspari.logging.utils import (
    _logging_extras,
    _logging_prefix,
    _logging_suffix,
    _resolve_environment,
)

"""
The steaming logs module is used to stream logs from an Azure Blob Storage container to the terminal.
Compared to standard Azure tooling, it provides a (much) lower latency.
"""


logger = logging.getLogger(__name__)


_default_conn_str_keys = [
    "AZURE_STORAGE_BLOB_URL",  # default Azure
    "ABS_CONNECTION_STRING",  # default Inspari
    "APP_ABS_CONNECTION_STRING",  # default Inspari App
    "API_ABS_CONNECTION_STRING",  # default Inspari Api
    "RAG_ABS_CONNECTION_STRING",  # default Inspari RAG
]


def _resolve_conn_str(
    conn_str: str | None, env_key: Optional[str] = None
) -> str | None:
    if conn_str is not None:
        return conn_str
    # If key is provided, use it.
    if env_key is not None:
        conn_str = os.getenv(env_key, None)
        if conn_str is None:
            logger.error(
                f"Failed to load connection string from (provided) environment variable {env_key}."
            )
            return None
    # Otherwise, try the default is priority order.
    for key in _default_conn_str_keys:
        conn_str = os.getenv(key, None)
        if conn_str is not None:
            return conn_str
    return None


def _get_blob_client(
    credential: DefaultAzureCredential,
    conn_str: Optional[str] = None,
    container: Optional[str] = None,
    blob: Optional[str] = None,
    env_key: Optional[str] = None,
    account_name: Optional[str] = None,
) -> BlobClient:
    # If container is not set, default to "logs" container.
    container = container if container is not None else "logs"
    # If blob is not set, use the default name.
    if account_name is not None:
        account_url = f"https://{account_name}.blob.core.windows.net"
        if blob is None:
            # List blobs in the container.
            blobs = (
                BlobServiceClient(account_url=account_url, credential=credential)
                .get_container_client(container)
                .list_blobs()
            )
            list_blobs = sorted(blobs, key=lambda x: x.name)
            # Get latest blob alphabetically.
            if len(list_blobs) == 0:
                blob = datetime.now(UTC).strftime("%Y-%m-%d.txt")
            else:
                blob = list_blobs[-1].name

        blob_client = BlobClient(
            account_url=account_url,
            container_name=container,
            blob_name=blob,
            credential=credential,
        )
        return blob_client
    # Get the connection string from the environment.
    if blob is None:
        now = datetime.now(UTC)
        blob = now.strftime("%Y-%m-%d.txt")
    conn_str = _resolve_conn_str(conn_str, env_key)
    if conn_str is None:
        raise ValueError("Failed to resolve connection string or account name.")
    return BlobClient.from_connection_string(conn_str, container, blob)


def _parse_bool(value: str | bool) -> bool:
    return value if isinstance(value, bool) else value.lower() == "true"


def _resolve_parameter(param: str | None, env_key: Optional[str] = None) -> str | None:
    if param is not None:
        return param
    if env_key is not None:
        for sub_key in env_key.split(","):
            value = os.getenv(sub_key, None)
            if value is not None:
                return value
    return None


class AzureBlobStorageHandler(logging.Handler):
    """
    Logging handler that streams logs to an Azure blob storage container.
    """

    def __init__(
        self,
        conn_str: Optional[str] = None,
        container: Optional[str] = None,
        blob: Optional[str] = None,
        env_key: Optional[str] = None,
        account_name_env_key: Optional[str] = None,
        account_name: Optional[str] = None,
        client_id: Optional[str] = None,
        client_id_env_key: Optional[str] = None,
        load_dot_env: bool | str = False,
        log_local: bool | str = False,
        exclude_environment_credentials: bool = False,
    ):
        self.client = None
        try:
            self.client = self._setup_client(
                conn_str,
                container,
                blob,
                env_key,
                account_name_env_key,
                account_name,
                client_id,
                client_id_env_key,
                load_dot_env,
                log_local,
                exclude_environment_credentials,
            )
        except Exception:
            logger.error("Failed to setup AzureBlobStorageHandler.")
            logger.error(traceback.format_exc())

    def _setup_client(
        self,
        conn_str: Optional[str] = None,
        container: Optional[str] = None,
        blob: Optional[str] = None,
        env_key: Optional[str] = None,
        account_name_env_key: Optional[str] = None,
        account_name: Optional[str] = None,
        client_id: Optional[str] = None,
        client_id_env_key: Optional[str] = None,
        load_dot_env: bool | str = False,
        log_local: bool | str = False,
        exclude_environment_credentials: bool = False,
    ) -> None | BlobClient:
        logging.Handler.__init__(self=self)
        # Check if running locally.
        local = _resolve_environment() == "local"
        if local and not _parse_bool(log_local):
            logger.debug("Not logging to Azure Blob Storage as environment is local.")
            return None
        # Setup credential.
        credential = DefaultAzureCredential(
            managed_identity_client_id=client_id,
            exclude_environment_credential=_parse_bool(exclude_environment_credentials),
        )
        # Optionally, load from .env file.
        if _parse_bool(load_dot_env):
            load_dotenv(credential=credential)
        # Resolve parameters.
        account_name = _resolve_parameter(account_name, account_name_env_key)
        # If we are local, we do NOT want to authenticate with a managed identity, so we skip the client_id.
        client_id = None if local else _resolve_parameter(client_id, client_id_env_key)
        # Setup client.
        client = _get_blob_client(
            credential, conn_str, container, blob, env_key, account_name
        )
        # Create blob if not exists.
        logger.info(f"Logging to storage account {client.account_name}.")
        if not client.exists():
            content_settings = ContentSettings(
                content_type="text/plain",
            )
            try:
                client.create_append_blob(content_settings=content_settings)
            except ResourceNotFoundError as e:
                logger.error(
                    f"Error creating blob: {e} {client.blob_name} in container {client.container_name} in {client.account_name}."
                )
                return
        logger.info(
            f"Logging to blob {client.blob_name} in container {client.container_name}."
        )
        return client

    def _prefix(self):
        pass

    def emit(self, record) -> None:
        if self.client is None:
            return
        msg = self.format(record)
        msg_with_extras = (
            f"{_logging_extras[_logging_prefix]}{msg}{_logging_extras[_logging_suffix]}"
        )
        self.client.append_block(f"{msg_with_extras}\n".encode("utf-8"))


def stream_logs(
    conn_str: Optional[str] = None,
    container: Optional[str] = None,
    blob: Optional[str] = None,
    env_key: Optional[str] = None,
    account_name: Optional[str] = None,
    account_name_env_key: Optional[str] = "APP_ABS_ACCOUNT_NAME",
    client_id_env_key: Optional[str] = None,
    client_id: Optional[str] = Field(
        None,
        description="The client ID for the managed identity. Not needed if using regular az cli login.",
    ),
    delay: int = 1,
):
    """
    Stream logs to terminal from the specified blob.
    """
    account_name = _resolve_parameter(account_name, account_name_env_key)
    client_id = _resolve_parameter(client_id, client_id_env_key)
    client = _get_blob_client(
        conn_str, container, blob, env_key, account_name, client_id=client_id
    )
    assert client is not None
    offset = 0
    first = True
    last_modified = None
    logging.info(
        f"Streaming logs from {client.blob_name} in container {client.container_name} in {client.account_name}."
    )
    while True:
        try:
            props = client.get_blob_properties()
            if last_modified is None or last_modified < props["last_modified"]:
                last_modified = props["last_modified"]
                bts = client.download_blob(offset=0).readall()
                offset += len(bts)
                print(bts.decode("utf-8"))
        except ResourceNotFoundError:
            if first:
                print("Blob not found.", end="")
            print(".", end="", flush=True)
        except HttpResponseError as e:
            print(f"Error: {e}")
        first = False
        time.sleep(delay)


def stream_logs_entrypoint():
    """
    Stream logs to terminal from the specified blob.
    """
    load_dotenv(dotenv_path=".env")
    # Set loglevel to info
    logging.basicConfig(level=logging.INFO)
    # Silence the azure logger
    logging.getLogger("azure.core.pipeline").setLevel(logging.WARNING)
    typer.run(stream_logs)


if __name__ == "__main__":
    stream_logs_entrypoint()
